/******************************
 MCL-C Translator
 Parser (mcl.y)
 
 GPL licensed - (c) 2006
 
                      3 marcos
******************************/

%{

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
//#define YYERROR_VERBOSE 1
#define YYDEBUG 1

#define BANNER "// Generated by MCL-C Translator"

/* struct */
typedef struct NODO {
	char dat[256];
	struct NODO *sin, *des;
} nodo;

typedef struct LIST {
	char var[256]; // nome variabile
	u_short in;
	struct LIST *next; // puntatore alla variabile successiva
} list;

typedef struct FUNCT_PROP {
	char funct_name[256]; //nome funzione
	int n_args; //numero degli argomenti della funzione
	struct FUNCT_PROP *next;
	list *var_list; //puntatore alla lista delle variabili locali della funzione i-esima
}funct_prop;

/* globals */
extern int num_linea;
nodo *head;
funct_prop *funct_list;
list *tmp_var_list=NULL;
char tmp_funct[256];

char buf[256];
int param_counter = 0;

/* protos */
nodo *mknodo (char [256], nodo *, nodo *);

int push_funct(char [256], int n);
int push_var(char dato[256], char f[256]);
int push_global_var(char dato[256], u_short in);
void clean_global_var();
int print_function_vars(char f[256]);

int check_var(char [256],char [256]);
int check_funct(char dato[256]);

int print_main (nodo *p);
int print_all (nodo *p);
int print_instr (nodo *p);
int print_call(nodo *p);
int print_esp(nodo *p);
int print_call_param(nodo *p);
int print_funct(nodo *p);
int print_if(nodo *p, u_short bit);
int print_cond (nodo *p);
int print_opc (nodo *p);
int print_while(nodo *p);

int yyerror (char *s);
%}

// definisco i tipi di token
%union {
  struct nodo *pnodo;
  char stringa[256];
}

%token FINELINEA OPLOGIC UGUALE PARENTESI_AP PARENTESI_CH EXIT MBEGIN END IF FI ELSE WHILE ELIHW PROGRAM OR AND NOT 
%token <stringa> NUMERO TEXT COMMENTO VIRGOLA FRASE PUNTOVIRGOLA OPMATE1 OPMATE2 OPLOGIC UGUALE GET

%type <pnodo> PARAM_L BODY M INSTR PARAM_L2 CL VL PL FL RET FUNCT CALL_FUNCT CALL_PARAM CALL_PARAM2 ESP T F COND A C PRED OPC PV

%start S
%%
S:	FINE_L CL VL PL M
	{
		head = mknodo("inizio",
				mknodo("globals",(nodo*) $2,(nodo*) $3),
				mknodo("function",(nodo *) $4, (nodo *) $5));
	}
;

CL:	/* empty */
	{ $$ = NULL;}
	| COMMENTO FINE_L CL
	{
	   sprintf(buf,"// %s",$1+1);
	$$ = (void *) mknodo("commento", mknodo(buf,NULL,NULL), (nodo *) $3);
	}
;

FINE_L:	/* empty */
	| FINELINEA FINE_L
;

VL:	/* empty */
	{
		$$ = NULL;
		push_funct("255",0);
	}
	| TEXT PUNTOVIRGOLA FINE_L CL VL
	{
	   sprintf(buf,"float %s;",$1);
	   
	 //in questo caso viene riutilizzato GLOBALS perch�la coppia CL VL �uguale a quella nell'assioma S...
   	   $$ = (void *) mknodo("glob", mknodo(buf,NULL,NULL), mknodo("globals",(nodo*) $4,(nodo*) $5));
	   
	   push_funct("255",0);

		if (!push_var($1, "255"))	// esiste gia' una variabile globale con lo stesso nome
		{	
			fprintf (stderr, "ERROR: global variable %s already declared. Abort!\n", $1);
			return 1;
		}	
	}
;

PL:	/* empty */
	{ $$ = NULL; }
	| PL FL CL
	{
		$$ = (void *) mknodo("prototipo",mknodo("prototipo2",(nodo *)$1, (nodo *)$2),(nodo*) $3);
	}
;

FL:	MBEGIN FUNCT BODY END FINE_L
	{		
	        if (push_funct(tmp_funct,param_counter)) // yeah
	        	$$ = (void *) mknodo("f_def",mknodo("id_f",mknodo("",NULL,NULL), (nodo *) $2),(nodo *) $3);
		else {
			$$ = NULL;
			fprintf (stderr, "ERROR: function %s already declared. Abort!\n", tmp_funct);
			return 1;
		}
		
		param_counter = 0;
		clean_global_var();
	}
	| MBEGIN RET FUNCT BODY END FINE_L
	{
		if (push_funct(tmp_funct,param_counter)) // yeah
		    $$ = (void *) mknodo("f_def",mknodo("id_f",(nodo *) $2, (nodo *) $3),(nodo *) $4);
		else {
			$$ = NULL;
			fprintf (stderr, "ERROR: function %s already declared. Abort!\n", tmp_funct);
			return 1;		
		}

		param_counter = 0;
		clean_global_var();
	}
;

RET:	TEXT UGUALE
	{
		sprintf(buf,"%s",$1);
		$$ = (void *) mknodo(buf,NULL,NULL);

		if (push_global_var ($1,0))	// return 1 ok
#ifdef MDEBUG
		fprintf (stderr, "DEBUG: parser RET: added new _var_ %s (0)\n", $1);
#endif
	}
;

FUNCT:	TEXT PARENTESI_AP PARAM_L PARENTESI_CH
	{
		sprintf(buf,"%s",$1);
		$$ = (void *) mknodo("f_name",mknodo(buf,NULL,NULL),(nodo *)$3);
		
		strcpy (tmp_funct,$1); // save function name
   	
	}
;

M:	MBEGIN PROGRAM PARENTESI_AP PARAM_L PARENTESI_CH BODY END FINE_L
	{
		$$ = (void *) mknodo("prog", (nodo *) $4, (nodo *) $6);	// sviluppiamo la lista parametri e il body
		
   		if (!push_funct("main",param_counter))
		   return 1;			      // error
		param_counter = 0;
	}
;

PARAM_L :	/* empty */
		{ $$ = NULL;}
		|  TEXT PARAM_L2
		{  
			sprintf(buf, "float %s",$1);
			$$ = (void *) mknodo(buf, (nodo *) $2 ,NULL);
			param_counter++;
			if (push_global_var($1,1))
#ifdef MDEBUG
			fprintf (stderr, "DEBUG: parser PARAM_L: added new _var_ %s (1)\n", $1);
#endif
		}
;

PARAM_L2:	/* empty */
		{ $$ = NULL;}
		| VIRGOLA TEXT PARAM_L2
		{
			sprintf(buf, "float %s",$2);
			$$ = (void *) mknodo(buf,(nodo *) $3,NULL);
			param_counter++;
			if (push_global_var($2,1))
#ifdef MDEBUG
			fprintf (stderr, "DEBUG: parser PARAM_L2: added new _var_ %s (1)\n", $2);
#endif
		}

;

BODY:	/* empty */
	{ $$ = NULL; }
	| FINE_L INSTR FINE_L BODY 
	{ $$ = (void *) mknodo("body", (nodo *) $2, (nodo *) $4); }
;

PV:	/* empty */
	{ $$ = NULL; }
	| PUNTOVIRGOLA
	{ $$ = (void *)mknodo("puntovirgola",NULL,NULL);	}
;

INSTR:	EXIT PV
	{
		strcpy(buf,"exit(1);");
		$$ = (void *) mknodo("exit",mknodo(buf,NULL,NULL),NULL);
	}
	| FRASE
	{
		char *s=malloc(strlen($1)-2);
		strncpy(s,$1+1,strlen($1)-2);
		sprintf(buf,"puts(\"%s\");",s);
		$$ = (void *) mknodo("frase",mknodo(buf,NULL,NULL),NULL);
	}
	| COMMENTO
	{ 
	   sprintf(buf,"// %s",$1+1);
	   $$ = (void *) mknodo("commento",mknodo(buf,NULL,NULL),NULL);
	}
	| CALL_FUNCT PV
	{
		$$ = (void *) mknodo("call_funct",(nodo *) $1, (nodo *) $2);
	}
	| RET ESP PV
	{
		$$ = (void *) mknodo("espr_rit",(nodo *) $1, mknodo("espr",(nodo *)$2, (nodo *)$3));
	}
	| ESP PV
	{
		$$ = (void *) mknodo("espr",(nodo *) $1, (nodo *) $2);
	}
	| TEXT GET
	{
		$$ = (void *) mknodo("get", mknodo ($1,NULL,NULL), NULL);
	}
	| IF COND BODY FI
	{
		$$ = (void *) mknodo ("if", (nodo *)$2, (nodo *)$3);	// sx=cond, dx=body
	}
	| IF COND BODY ELSE BODY FI
	{
		$$ = (void *) mknodo ("if_fi", (nodo *)$2, mknodo ("else",(nodo *)$3, (nodo *)$5));
					       // sx=cond,                dx=body,     body
	}
	| WHILE COND BODY ELIHW
	{
		$$ = (void *) mknodo("while",(nodo *)$2, (nodo *)$3);
	}
;

CALL_FUNCT:	TEXT PARENTESI_AP CALL_PARAM PARENTESI_CH
		{
			$$ = (void *) mknodo("call",mknodo($1,NULL,NULL),(nodo *) $3);
		}
;

CALL_PARAM:	/* empty */
		{ $$ = NULL }
		| ESP CALL_PARAM2
		{
			$$ = (void *) mknodo("call_1",(nodo *) $1,(nodo *) $2);
		}
;

CALL_PARAM2:	/* empty */
		{ $$ = NULL }
		| VIRGOLA ESP CALL_PARAM2
		{
			$$ = (void *) mknodo("call_2",(nodo *) $2,(nodo *) $3);
		}
;

ESP:	ESP OPMATE1 T
	{
		$$ = (void *) mknodo("esp",(nodo *) $1, mknodo("operatore",mknodo($2,NULL,NULL),(nodo *) $3));
	}
	| OPMATE1 T
	{
		$$ = (void *) mknodo("esp",NULL , mknodo("operatore",mknodo($1,NULL,NULL),(nodo *) $2));
	}
	| T
	{
		$$ = (void *) mknodo("termine",(nodo *) $1,NULL);
	}
;

T:	T OPMATE2 F
	{
		$$ = (void *) mknodo("esp2",(nodo *) $1, 
		mknodo("2nd_fattore",mknodo($2,NULL,NULL),(nodo *) $3));
	}
	| F
	{
		$$ = (void *) mknodo("fattore",(nodo *) $1, NULL);
	}
;

F:	CALL_FUNCT

	{
		$$ = (void *) mknodo("call_funct",(nodo *) $1, NULL);
	}
	| NUMERO
	{
		$$ = (void *) mknodo("numero",mknodo($1,NULL,NULL),NULL);
	}
	| TEXT
	{
		$$ = (void *) mknodo("testo",mknodo($1,NULL,NULL),NULL);

		if (push_global_var ($1,0))	// return 1 ok
#ifdef MDEBUG
		fprintf (stderr, "DEBUG: parser ESP: added new _var_ %s (0)\n", $1);
#endif

	}
	| PARENTESI_AP ESP PARENTESI_CH
	{
		$$ = (void *) mknodo("espressione",(nodo *) $2,NULL);
	}
;

COND:	PARENTESI_AP COND OR A PARENTESI_CH
	{
		$$ = (void *) mknodo("condOR",(nodo*) $2, (nodo *) $4);
	}
	| A
	{	
		$$ = (void *) mknodo("condA",(nodo*) $1,NULL);
	}

;

A:	PARENTESI_AP A AND C PARENTESI_CH
	{
		$$ = (void *) mknodo("condAND",(nodo *) $2, (nodo *) $4);
	}
	| C
	{	
		$$ = (void *) mknodo("condC",(nodo *)$1,NULL);
	}
;

C:	F
	{
		$$ = (void *) mknodo("f",(nodo *)$1,NULL);
	}
	| COND
	{
		$$ = (void *) mknodo("cond_2",(nodo *)$1,NULL);
	}
	| PARENTESI_AP NOT COND PARENTESI_CH
	{	$$ = (void *) mknodo("cond_not",(nodo *) $3,NULL);	}
	| PARENTESI_AP PRED PARENTESI_CH
	{	$$ = (void *) mknodo("condPRED",(nodo*)$2,NULL);	}
;

PRED:	ESP OPC ESP
	{	$$=(void *) mknodo("pred",(nodo*)$2,mknodo("pred-esp",(nodo*)$1,(nodo*)$3));	}
;

OPC:	OPLOGIC
	{	$$=(void *) mknodo("oplogic",mknodo($1,NULL,NULL),NULL);	}
	| UGUALE
	{	$$=(void *) mknodo("uguale",mknodo($1,NULL,NULL),NULL);		}
;

	
%%

nodo *mknodo (char d[256], nodo *sx, nodo *dx)
{
	nodo *temp = (nodo*) malloc(sizeof(nodo));
	if (temp==NULL)
	{
		fprintf (stderr, "ERROR: malloc() on nodo\n");
		exit (1);
	}

	strncpy(temp->dat, d, strlen(d)*sizeof(char));
	temp->sin = sx;
	temp->des = dx;
	
	return (temp);
}

// ok
int push_funct(char dato[256],int n)
{
#ifdef MDEBUG
fprintf (stderr, "DEBUG: parser push_funct(): adding function %s with #param = %d\n", dato, n);
#endif
	
	if (!check_funct(dato)) // non trovato
	{
		funct_prop *new_node = (funct_prop *) malloc(sizeof (funct_prop));
		
		strcpy(new_node->funct_name, dato);
		new_node->n_args = n;
		
		if (strcmp(dato,"255") && tmp_var_list) {	// non e' la 255 ed ha parametri
			// copio tutta la lista delle variabili tmp_var_list a var_list
			
			list *p = tmp_var_list;
			
			while (p!=NULL) {
				list *tmp = (list *) malloc(sizeof(list));
				strcpy (tmp->var, p->var);
				tmp->in=p->in;
				tmp->next=new_node->var_list;
				new_node->var_list=tmp;
				
				p=p->next;
			}
		
		}
		else
			new_node->var_list=NULL;
		
		new_node->next = NULL;
		
		if (funct_list == NULL)	// la lista delle funzioni e' vuota
			funct_list = new_node;
		else			// abbiamo roba dentro
		{
			funct_prop *p = funct_list;
			while (p->next != NULL)	// andiamocene alla fine
				p = p->next;
			p->next = new_node;
		}
		
#ifdef MDEBUG
	fprintf(stderr, "DEBUG: parser push_funct(): [OK] added function %s\n", dato);
#endif

      		return 1;	// ok
	}
	else
	{
#ifdef MDEBUG
	if (strcmp(dato,"255"))
		fprintf(stderr, "DEBUG: parser push_funct(): [KO] function %s already present ...\n",dato);
#endif

      		return 0;	// ko
	}
	
	return 0; 
}


int push_var(char dato[256], char f[256])
{
	funct_prop *p = funct_list;
	list *t, *t_p;
	u_short funct_found=0;
	
	/* andiamo a prenderci la funzione che ci interessa e la scorriamo */
	while (p!=NULL)
	{
		if (!strcmp(p->funct_name,f))
		{
			funct_found=1;
			t=p->var_list;	// eccola qua la lista delle mie funzioni
			break;
		}
		
		p=p->next;
	}
	if (!funct_found)
	{
#ifdef MDEBUG
	fprintf(stderr, "DEBUG: parser push_var(): function %s not found while pushing var. Abort\n", f);
#endif
		return 0;	// error
	}	
	
	// vediamo se c'e' gia' in lista 
	t_p=t;
	while (t_p!=NULL) 
	{
		if (!strcmp(t_p->var, dato)) {
#ifdef MDEBUG
			fprintf(stderr, "DEBUG: parser push_var(): var %s alreay in function, Abort\n", dato);
#endif
			return 0;	// error
		}
		t_p=t_p->next;
	}
	
	// mallochiamo
	list *new_node = (list *) malloc(sizeof(list));
	strcpy(new_node->var, dato);
	
	//inserimento in testa
	new_node->next = t;
	p->var_list=new_node;	
	
#ifdef MDEBUG
	fprintf (stderr, "DEBUG: parser push_var(): [OK] added %s in %s\n", new_node->var, p->funct_name);
#endif
	
	return 1;
}

// push the var in tmp_var_list
int push_global_var(char dato[256], u_short in) 
{
	// vediamo se e' gia' in lista
	if (tmp_var_list!=NULL) 
	{
		list *p=tmp_var_list;
		while (p!=NULL) {
			if (!strcmp(p->var,dato))
				return 0;
			p=p->next;
		} 
	}
/********** Gestione Doppione variabili globali/locali ******************/
		
	if (in == 0) //se la variabile �locale, verifico se prima �stata definita tra le variabili globali
	{
		list *temp = funct_list->var_list;
		while (temp!= NULL)
		{
			if (!strcmp(temp->var,dato))
				return 0;
			temp=temp->next;
		}
	}
	
/*************************************************************************/
	// ok inseriamo
	list *new_node = (list *) malloc(sizeof(list));
	strcpy(new_node->var, dato);
	new_node->in=in;		// tipo di variabile -> 1=data in input alla funzione, 0 definita localmente
	new_node->next = tmp_var_list;
	tmp_var_list=new_node;
	
	return 1;	// ok
}

void clean_global_var() {
	if (tmp_var_list) {
		free (tmp_var_list);
		tmp_var_list=NULL;
	}
}

// ok
int check_funct(char dato[256])
{
	funct_prop *p=funct_list;
	
#ifdef MDEBUG
	fprintf(stderr, "DEBUG: parser check_funct(): checking _function_ %s in the list\n", dato);
#endif
	
	while (p != NULL)
	{		
		if (!strcmp(p->funct_name, dato)){
#ifdef MDEBUG
		fprintf(stderr, "DEBUG: parser check_funct(): [OK] function found\n");
#endif
			return 1;
		}
		else
			p = p->next;
	}
	
#ifdef MDEBUG
fprintf(stderr, "DEBUG: parser check_funct(): [KO] function not found!\n");
#endif
	
	return 0;
}

int check_var(char dato[256], char f[256])
{
	funct_prop *p=funct_list;	// dobbiamo scorrere tutta la lista delle funzioni
	
	int funct_found = 0;
	
#ifdef MDEBUG
fprintf(stderr, "DEBUG: parser check_var(): checking function %s for _var_ %s\n", f, dato);
#endif
	
	/*
	 prima vediamo se e' globale!
	*/
	list *t = funct_list[0].var_list;
	
	// variabili globali
	while (t!=NULL)
	{
		if (!strcmp(t->var, dato))
		{
#ifdef MDEBUG
		fprintf(stderr, "DEBUG: parser check_var(): variable %s found as GLOBAL\n", dato);
#endif
			return 1;
		}
		
		t=t->next;
	}
	
	/* ora andiamo a prenderci la funzione che ci interessa e la scorriamo */
	while (p!=NULL)
	{
		if (!strcmp(p->funct_name,f))
			{
				funct_found=1;
				break;
			}
		else
			p=p->next;
	}
	
	if (!funct_found)
	{
#ifdef MDEBUG
	fprintf(stderr, "DEBUG: parser check_var(): function %s not found while searching for var. Abort\n", f);
#endif
		return 0;	// error
	}
	
	// variabli della funzione
	t = p->var_list;
	while (t!=NULL)
	{
		if (!strcmp(t->var, dato))
		{
#ifdef MDEBUG
		fprintf(stderr, "DEBUG: parser check_var(): variable found\n");
#endif
			return 1;
		}
		t=t->next;
	}
	
#ifdef MDEBUG
fprintf(stderr, "DEBUG: parser check_var(): variable not found\n");
#endif
	
	return 0; 	// error
}

/* scrive all'inizio della funzione f le sue variabili */
int print_function_vars(char f[256])
{
	funct_prop *p = funct_list;
	list *t;
	u_short funct_found=0;
	
	/* andiamo a prenderci la funzione che ci interessa e la scorriamo */
	while (p!=NULL)
	{
		if (!strcmp(p->funct_name,f))
		{
			funct_found=1;
			t=p->var_list;	// eccola qua la lista delle mie funzioni
			break;
		}	
		p=p->next;
	}
	if (!funct_found)
	{
#ifdef MDEBUG
	fprintf(stderr, "DEBUG: parser print_function_vars(): function %s not found while\n", f);
#endif
		return 0;	// error
	}
	
	printf ("// function variables\n");
	while (t!=NULL) {
		if (!t->in) 	// sempre che non sia di input :)
			printf ("float %s;\n", t->var); 
		t=t->next;
	}

	return 1;
}

int print_all(nodo *p)
{	
	puts (BANNER);
	printf("//--------------------\n");
	printf("\n");
	printf("#include <stdlib.h>\n");
	printf("#include <stdio.h>\n");
	printf("\n");
	
	// stampiamo le variabili globali
	print_function_vars ("255");
	
	// diamo il via alle danze
	if (print_instr(p->des))
		return 1;

	return 0;	
}

int print_main(nodo *p)
{
	u_short i=1;
	
	// print dei parametri
	if (p->sin==NULL)
		printf ("\nint main ()\n{\n");
	else
		{
			printf ("\nint main (int argc, char **argv)\n{\n");
	        	nodo *temp = p;
			while(temp->sin != NULL)
				{
					temp = (nodo*) temp->sin;
					printf("%s = atof (argv[%d]);\n", temp->dat, i);
					i++;
				}
		}

	print_function_vars("main");
	print_instr(p->des); // body

	printf("\nreturn 0;\n");
	printf("}\n\n");
	printf("// ==EOF==\n");
	
	return 0;
}

int print_funct(nodo *p)
{
	if (p!= NULL)
	{
		if (strcmp(p->sin->sin->dat,"")==0) //se vero allora la funzione �di tipo void
			printf("void ");
		else
			printf("float ");


		nodo *temp=p->sin;
		printf("%s (",temp->des->sin->dat);
		
		strcpy(buf,temp->des->sin->dat);
		
		temp = temp->des->des;
		
		while(temp != NULL)
		{
			printf("%s",temp->dat);
			temp = (nodo*) temp->sin;
			if (temp!=NULL)
				printf(",");
		}
		printf(")\n{\n");
	
		// stampiamo le variabili della funzione
		print_function_vars(p->sin->des->sin->dat);
		
		//print BODY
		temp = p->des;
			print_instr(temp);
		free(temp);
		if (strcmp(p->sin->sin->dat,"")!=0)
			printf("return (%s);\n",p->sin->sin->dat);
		printf("}\n\n");
	}
	return 0;
}

int print_call(nodo *p)
{
	//stampo nome chiamata
	printf("%s(",p->sin->dat);

	int param_expected=-1;
	funct_prop *temp = funct_list;
	
	/* andiamo a prenderci la funzione che ci interessa e la scorriamo */
	while (temp!=NULL)
	{
		if (!strcmp(temp->funct_name,p->sin->dat))
		{
			param_expected = temp->n_args;
			break;
		}	
		temp=temp->next;
	}
	
	if (param_expected == -1)
	{
		fprintf (stderr,"\nERROR: function \"%s\" not found\n",p->sin->dat);
		exit (1);	
	}
	//stampo parametri
	param_counter=0;
	if (p->des) 	// c'e' almeno un parametro
		param_counter = print_call_param(p->des) +1;
	
	//chiudo la chiamata
	printf(")");
	
	if (param_counter != param_expected)
	{
		fprintf (stderr,"\nERROR: Number of parameters in function \"%s\" exceed with definition",p->sin->dat);
		fprintf (stderr,"\nERROR: Expecting  %d parameters instead of %d\n",param_expected,param_counter);
		exit(1);
	}

	return 0;
}

// ok
// stampa i parametri dalla funzione e verifica se son in numero giusto
int print_call_param(nodo *p)
{
	if(p != NULL)
	{
		if (strcmp(p->dat,"call_2")==0){
			printf(", ");
			param_counter++;
			print_call_param(p->sin);
			print_call_param(p->des);
		}
		else if (strcmp(p->dat,"operatore")==0){
			printf("%s",p->sin->dat);
			print_call_param(p->des);
		}
		else if (strcmp(p->dat,"2nd_fattore")==0){
			printf("%s",p->sin->dat);
			print_call_param(p->des);
		}
		else if (strcmp(p->dat,"numero")==0){
			print_esp(p);
		}
		else if (strcmp(p->dat,"testo")==0){
			print_esp(p);
		}
		else if (strcmp(p->dat,"call_funct")==0){
			print_esp(p);
		}
		else if (strcmp(p->dat,"espressione")==0){
			printf("(");
			print_call_param(p->sin);
			printf(")");
		}
		else {
			print_call_param(p->sin);
			print_call_param(p->des);
		}
	}
	return param_counter;
}



int print_esp(nodo *p)
{	
	if (p!=NULL)
	{
#ifdef MDEBUG
	fprintf (stderr,"DEBUG: parser print_esp(): node: %s\n", p->dat);
#endif
		
		
		if (strcmp(p->dat,"numero")==0){
			printf("%s",p->sin->dat);
		}
		else if (strcmp(p->dat,"testo")==0) {
			printf("%s",p->sin->dat);
		}
		else if (strcmp(p->dat,"call_funct")==0){
			print_instr(p->sin);
		}
		else if (strcmp(p->dat,"operatore")==0){
			printf("%s",p->sin->dat);
			print_esp(p->des);
		}
		else if (strcmp(p->dat,"2nd_fattore")==0){
			printf("%s",p->sin->dat);
			print_esp(p->des);
		}
		else if (strcmp(p->dat,"espressione")==0){
			printf("( ");
			print_esp(p->sin);
			printf(" )");
		}
		else
		{
			print_esp(p->sin);
			print_esp(p->des);
		}
	}
	return 0;
}

int print_if(nodo *p, u_short bit) {
	/* bit = 0 -> WITHOUT else
	   bit = 1 -> WITH else */
	
	nodo *temp;
	
	printf ("if ");
	
	/* sviluppo il COND */
	print_cond(p->sin);
	printf ("\n");
	
		
	/* print del 1st BODY */
	printf ("{\n");
	if (!bit)	temp = p->des;
	else		temp = p->des->sin;
	
	if (print_instr(temp))
		return 1;
	printf ("}\n");
	
	/* print del 2nd BODY (solo se c'e' l'else)*/
	if (bit) 
	{
		printf ("else {\n");
		temp = p->des->des;
		print_instr(temp);
		printf ("}\n");
	}		
	
	if (temp)
		free(temp);	
	
	return 0;
}

int print_cond (nodo *p) {
	if (p == NULL)
		return 1;	// error

#ifdef MDEBUG
fprintf (stderr,"DEBUG: parser print_cond(): node: %s\n", p->dat);
#endif
	
	if (!strcmp (p->dat, "condA"))
		print_cond (p->sin);
		
	else if (!strcmp (p->dat, "condC"))
		print_cond (p->sin);
	
	else if (!strcmp (p->dat, "condPRED"))
	{
		printf("( ");
		print_cond (p->sin);
		printf(" )");
	}
	
	else if (!strcmp (p->dat, "f"))
		print_esp(p->sin);
	
	else if (!strcmp (p->dat,"condOR"))
	{
		printf("( ");
		print_cond(p->sin);
		printf(" || ");
		print_cond(p->des);
		printf(" )");
	}
	else if (!strcmp (p->dat,"condAND"))
	{
		printf("( ");
		print_cond(p->sin);
		printf(" && ");
		print_cond(p->des);
		printf(" )");
	}

	else if (!strcmp(p->dat,"cond_2"))
	{
		print_cond(p->sin);
	}
	else if (!strcmp(p->dat,"cond_not"))
	{
		printf("( !");
		print_cond(p->sin);
		printf(" )");
	}
	
	else if (!strcmp (p->dat, "pred"))
	{
		print_esp(p->des->sin);
		if (print_opc (p->sin))
			fprintf (stderr, "ERROR: unknown error in print_cond\n");
		print_esp(p->des->des);
	}
	else
		return 1;
	
	return 0;
}

int print_opc (nodo *p) {
	if (p == NULL)	
		return 1;	// error
	
	if (!strcmp (p->dat, "oplogic"))
		printf ("%s", p->sin->dat);
	else if (!strcmp (p->dat, "uguale"))
		printf ("==");
	else
	{	
		fprintf (stderr,"ERROR: %s unknown logic operator\n", p->dat);
		return 1;
	}
	
	return 0;
}

int print_while(nodo *p)
{
	printf("\nwhile");
	print_cond(p->sin);
	printf("\n{\n");
	print_instr(p->des);
	printf("\n}\n");
	return 0;	
}
int print_instr(nodo *p) {
 if (p != NULL)
	{
	
#ifdef MDEBUG
	fprintf (stderr, "DEBUG: parser print_instr(): %s\n", p->dat);
#endif

	if (strcmp(p->dat,"prog")==0)  // main
		print_main(p);

	else if (strcmp(p->dat,"commento")==0)  //  commento
	{
		printf("%s\n",p->sin->dat);
		print_instr(p->des);
	}

	else if (strcmp(p->dat,"glob")==0)  // variabili globali
	{
		printf("%s\n",p->sin->dat);
		print_instr(p->des);
	}
	
	else if (strcmp(p->dat,"f_def")==0)  // definizione di una funzione
		print_funct(p);

	else if (strcmp(p->dat,"exit")==0)  // exit
	{
		printf("%s\n",p->sin->dat);
		print_instr(p->des);
	}

	else if (strcmp(p->dat,"frase")==0)  // frase
	{
		printf("%s\n",p->sin->dat);
		print_instr(p->des);
	}
		
	else if (!strcmp(p->dat,"get"))
	{
		printf("scanf(\"%%f\", ");
		printf("&%s",p->sin->dat);
		printf(");\n");
	}
	
	else if (strcmp(p->dat,"espr")==0)
	{
		if (p->des == NULL)
		{
			printf("printf(\"%%.4f\\n\", (float)");
			print_esp(p->sin);
			printf(");\n");
		}
		else
		{
		 	/* salviamo il nome della funzione per vedere nel
			print_esp se la variabile e' gia' stata dichiarata */
			strcpy (tmp_funct,p->sin->dat);
			
			print_esp(p->sin);
			printf(";\n");
		}	
	}
	
	else if (strcmp(p->dat,"espr_rit")==0)
	{
		printf("%s = ",p->sin->dat);
		print_esp(p->des->sin);
		printf(";\n");
		
		if (p->des->des == NULL)
		{
			printf("printf(\"\\n  %s =\\n      %%.4f\\n\\n\",",p->sin->dat);
			printf("%s",p->sin->dat);
			printf(");\n");
		}
	}
	
	else if (strcmp(p->dat,"call_funct")==0)
	{
		if (p->des == NULL)
		{
			printf("printf(\"%%.4f\\n\",");
			print_instr(p->sin);
			printf(");\n");
		}
		else
		{
			print_instr(p->sin);
			printf(";\n");
		}	

	}

	else if (strcmp(p->dat,"call")==0)  // chiamata di funzione
		print_call(p);

	else if (!strcmp (p->dat, "if"))
		print_if(p, 0);			// 0 WITHOUT else
	
	else if (!strcmp (p->dat, "if_fi"))
		print_if(p, 1);			// 1 WITH else
				
	else if (!strcmp (p->dat, "while"))
		print_while(p);
	
	else
	{
		print_instr(p->sin);
		print_instr(p->des);
	}

	
	return 0;
	}
 else
	return 1;
}


int yyerror (char *s)
{
	fprintf(stderr, "*** ERROR: parser: %s ... at line %d\n",s,num_linea);
	return (1); 
}

int main(int argc, char **argv)
{
	u_short ret;
#ifdef DEBUG
	yydebug = 1;
#endif

#ifdef MDEBUG
	fprintf (stderr, "DEBUG ENABLED with -DMDEBUG\n");
#endif


	ret=yyparse();

	if (ret)
		return (ret);
	else
		return (print_all(head));
}
